#coding: utf-8
try:
    from collections import OrderedDict
except ImportError:
    # Python 2.6 alternative
    from ordereddict import OrderedDict


class Dictionary(OrderedDict):
    """
    Dictionary that exposes keys as properties for easy read access.
    """
    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError("%s without attr '%s'" %
                             (type(self).__name__, name))



#coding: utf-8
# from __future__ import unicode_literals
import re
import weakref
import collections
import logging

from .cache import Cache
from .item import Dictionary

logger = logging.getLogger(__name__)


class ParserMixin(object):
    """
    A mixin that holds all base parser implementation.
    """

    number_of_blocks_in_cache = 0
    default_item_class = Dictionary
    yield_item_to_each_field_value_found = False
    begin = None
    end = None

    def __init__(self):
        self.begin_found = not self.has_search_begin
        self.cache = Cache(self.number_of_blocks_in_cache + 1)
        self._assign_parser_to_fields()

    def _assign_parser_to_fields(self):
        """
        Assigns an weak parser reference to fields.
        """
        ref = weakref.ref(self)
        for item in list(self.fields.values()):
            if hasattr(item, 'assign_parser'):
                item.assign_parser(ref)

    def parse(self, iterator):
        for item in self.parse_iterator(iterator):
            yield item

    @property
    def has_item(self):
        return hasattr(self, 'item') and self.item is not None

    def parse_iterator(self, iterator):
        try:
            while True:
                block = next(iterator)
                res = self.parse_block(block)
                if res:
                    yield res
        except StopIteration:
            res = self.finalize()
            if res:
                yield res

    def parse_block(self, block):
        logger.debug('parse_block: %r:%s', type(block), block)
        self.cache.append(block)

        if self.has_search_begin and not self.begin_found:
            self.begin_found = bool(self._begin.match(block))

        if self.begin_found:
            logger.debug('init found: %r', self.begin_found)
            if not self.has_item:
                self.item = self.default_item_class()
            if self.has_search_end:
                self.begin_found = not bool(self._end.match(block))

            for block in self.cache.consume():
                for name, field in list(self.fields.items()):
                    if name in self.item and \
                            hasattr(field, 'is_list') and not field.is_list:
                        continue
                    value = field.parse_block(block)
                    if value is not None:
                        self.assign_value_into_item(name, value)
                        if self.yield_item_to_each_field_value_found:
                            return self.finalize_item()

            if not self.begin_found:
                return self.finalize_item()

    def finalize(self):
        if not self.has_item:
            return None
        if self.yield_item_to_each_field_value_found:
            return None
        return self.finalize_item()

    def finalize_item(self):
        for name, field in list(self.fields.items()):
            if not name in self.item:
                value = None
                if hasattr(field, 'finalize') and \
                        isinstance(field.finalize, collections.Callable):
                    value = field.finalize()
                if value is None and hasattr(field, 'default'):
                    value = field.default
                if value is not None:
                    self.assign_value_into_item(name, value)

        res = self.process_item(self.item)
        self.item = None
        return res

    def assign_value_into_item(self, name, value):
        if isinstance(value, list) and not name in self.item:
            self.item[name] = value
        elif isinstance(value, list) and hasattr(self.item[name], 'extend'):
            self.item[name].extend(value)
        else:
            self.item[name] = value

    def process_item(self, item):
        "Allows final modifications at the object being returned"
        return item


class ParserMetaclass(type):
    """
    Collect data-extractors into a field collection and injects ParserMixin.
    """
    def __new__(self, name, bases, attrs):
        if object in bases:
            bases = tuple([c for c in bases if c != object])

        return type.__new__(self, name, bases + (ParserMixin,), attrs)

    def __init__(cls, name, bases, attrs):
        super(ParserMetaclass, cls).__init__(name, bases, attrs)

        cls.fields = dict((k, v) for k, v in list(attrs.items())
                          if hasattr(v, 'parse_block')
                          and not isinstance(v, type))

        cls.add_regex_attr(cls, attrs, 'begin')
        cls.add_regex_attr(cls, attrs, 'end')

        for name, attr in list(cls.fields.items()):
            if hasattr(attr, 'assign_class'):
                attr.assign_class(cls, name)

    def add_regex_attr(self, cls, attrs, name):
        has_attr = name in attrs
        setattr(cls, 'has_search_'+name, has_attr)
        if has_attr:
            regex = attrs[name]
            logger.error('add_regex_attr %s: %s', name, regex)
            setattr(cls, '_' + name, re.compile(regex, re.UNICODE))


Parser = ParserMetaclass('Parser', (object,), {})



#coding: utf-8
from collections import deque


class Cache(object):
    def __init__(self, max_length=0):
        self.max_length = max_length
        self.items = deque()

    def __len__(self):
        return len(self.items)

    def append(self, item):
        self.items.append(item)
        if self.max_length:
            while len(self.items) > self.max_length:
                self.items.popleft()

    def itens(self):
        return self.items

    def consume(self):
        while self.items:
            yield self.items.popleft()



#coding: utf-8

"""

Fields define how and what data will be extracted. The parser does not expect
the fields explicitly inherit from :py:class:`~raspador.fields.BaseField`, the
minimum expected is that a field has at least a method `parse_block`.

The fields in this file are based on regular expressions and provide conversion
for primitive types in Python.
"""

import re
from datetime import datetime
import collections


class BaseField(object):
    """
    Contains processing logic to extract data using regular expressions, and
    provide utility methods that can be overridden for custom data processing.


    Default behavior can be adjusted by parameters:

    search

        Regular expression that must specify a group of capture. Use
        parentheses for capturing::

            >>> s = "02/01/2013 10:21:51           COO:022734"
            >>> field = BaseField(search=r'COO:(\d+)')
            >>> field.parse_block(s)
            '022734'

        The `search` parameter is the only by position and hence its name can
        be omitted::

            >>> s = "02/01/2013 10:21:51           COO:022734"
            >>> field = BaseField(r'COO:(\d+)')
            >>> field.parse_block(s)
            '022734'


    input_processor

        Receives a function to handle the captured value before being returned
        by the field.

            >>> s = "02/01/2013 10:21:51           COO:022734"
            >>> def double(value):
            ...     return int(value) * 2
            ...
            >>> field = BaseField(r'COO:(\d+)', input_processor=double)
            >>> field.parse_block(s)  # 45468 = 2 x 22734
            45468

    groups

        Specify which numbered capturing groups do you want do process in.


        You can enter a integer number, as the group index::

            >>> s = "Contador de Reduções Z:                     1246"
            >>> field = BaseField(r'Contador de Reduç(ão|ões) Z:\s*(\d+)', \
                groups=1, input_processor=int)
            >>> field.parse_block(s)
            1246

        Or a list of integers::

            >>> s = "Data do movimento: 02/01/2013 10:21:51"
            >>> c = BaseField(r'^Data .*(movimento|cupom): (\d+)/(\d+)/(\d+)',\
                groups=[1, 2, 3])
            >>> c.parse_block(s)
            ['02', '01', '2013']

        .. note::

            If you do not need the group to capture its match, you can optimize
            the regular expression putting an `?:` after the opening
            parenthesis::

            >>> s = "Contador de Reduções Z:                     1246"
            >>> field = BaseField(r'Contador de Reduç(?:ão|ões) Z:\s*(\d+)')
            >>> field.parse_block(s)
            '1246'

    default

        If assigned, the :py:class:`~raspador.parser.Parser` will query this
        default if no value was returned by the field.

    is_list

        When specified, returns the value as a list::

            >>> s = "02/01/2013 10:21:51           COO:022734"
            >>> field = BaseField(r'COO:(\d+)', is_list=True)
            >>> field.parse_block(s)
            ['022734']

        By convention, when a field returns a list, the
        :py:class:`~raspador.parser.Parser` accumulates values
         returned by the field.

    """
    def __init__(self, search=None, default=None, is_list=False,
                 input_processor=None, groups=[]):
        self.search = search
        self.default = default
        self.is_list = is_list
        self.input_processor = input_processor
        self.groups = groups

        if self.input_processor and \
                not isinstance(self.input_processor, collections.Callable):
            raise TypeError('input_processor is not callable.')

        if not hasattr(self.groups, '__iter__'):
            self.groups = (self.groups,)

        self._setup()

    @property
    def _search_method(self):
        return self.search.findall

    def _setup(self):
        "Hook to special setup required on child classes"
        pass

    def assign_parser(self, parser):
        """
        Receives a weak reference of
        :py:class:`~raspador.parser.Parser`
        """
        self.parser = parser

    def _is_valid_result(self, value):
        return bool(value)

    def _process_value(self, value):
        if self.groups:
            if len(value) == 1:   # take first, if only one item
                value = value[0]
            length = len(value)
            value = [value[i] for i in self.groups if i < length]

        if len(value) == 1:   # take first, if only one item
            value = value[0]
        return value

    def to_python(self, value):
        """
        Converts parsed data to native python type.
        """
        return value

    @property
    def search(self):
        return self._search

    @search.setter
    def search(self, value):
        self._search = re.compile(value, re.UNICODE) if value else None

    def assign_class(self, cls, name):
        self.cls = cls

    def parse_block(self, block):
        if self.search:
            value = self._search_method(block)
            if self._is_valid_result(value):
                value = self._process_value(value)
                value = self.to_python(value)
                if self.input_processor:
                    value = self.input_processor(value)
                if value is not None and self.is_list \
                        and not isinstance(value, list):
                    value = [value]
                return value


class StringField(BaseField):
    def to_python(self, value):
        return str(value).strip()


class FloatField(BaseField):
    "Removes thousand separator and converts to float."
    def to_python(self, value):
        value = value.replace(',', '')
        return float(value)


class BRFloatField(BaseField):
    "Removes thousand separator and converts to float (Brazilian format)"
    def to_python(self, value):
        value = value.replace('.', '')
        value = value.replace(',', '.')
        return float(value)


class IntegerField(BaseField):
    def to_python(self, value):
        return int(value)


class BooleanField(BaseField):
    """
    Returns true if the block is matched by Regex, and is at least some value
    is captured.
    """
    def _setup(self):
        self.default = False

    @property
    def _search_method(self):
        return self.search.match

    def _process_value(self, value):
        res = value.groups() if value else False
        return super(BooleanField, self)._process_value(res)

    def _is_valid_result(self, value):
        return value and (value.groups())

    def to_python(self, value):
        return bool(value)


class DateField(BaseField):
    """

    Field that holds data in date format, represented in Python by
    datetine.date.

        http://docs.python.org/library/datetime.html
    """

    default_format_string = '%d/%m/%Y'
    convertion_function = lambda self, date: datetime.date(date)

    def __init__(self, search=None, format_string=None, **kwargs):
        self.format_string = format_string \
            if format_string else self.default_format_string
        super(DateField, self).__init__(search=search, **kwargs)

    def to_python(self, value):
        date_value = datetime.strptime(value, self.format_string)
        return self.convertion_function(date_value)


class DateTimeField(DateField):
    """
    Field that holds data in hour/date format, represented in Python by
    datetine.datetime.

        http://docs.python.org/library/datetime.html
    """

    default_format_string = '%d/%m/%Y %H:%M:%S'
    convertion_function = lambda self, date: date


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# flake8: noqa

from .parser import Parser
from .item import Dictionary
from .fields import BaseField, StringField, FloatField, BRFloatField, \
    IntegerField, DateField, DateTimeField, BooleanField

from .decorators import FieldProxy, UnionUntilRegexProxy

from .cache import Cache



#coding: utf-8
import re


class FieldProxy(object):
    def __init__(self, field):
        self.field = field

    def __getattr__(self, attr):
        return getattr(self.field, attr)


class UnionUntilRegexProxy(FieldProxy):
    """
    Does cache of blocks until the provided regex returns a match, then uses
    the ``union_method`` to join blocks that are sent to the decorated field.
    """
    def __init__(self, field, union_method, search_regex):
        super(UnionUntilRegexProxy, self).__init__(field)
        self.cache = []
        self.union_method = union_method
        self.search_regex = re.compile(search_regex, re.UNICODE)

    def parse_block(self, block):
        if hasattr(block, 'rstrip'):
            block = block.rstrip()
        self.cache.append(block)
        if self.search_regex.match(block):
            blocks = self.union_method(self.cache)
            self.cache = []
            return self.field.parse_block(blocks)



#!/usr/bin/env python
#coding: utf-8

from distutils.core import setup

with open('README.rst') as f:
    long_description = f.read()

setup(
    name='raspador',
    author='Fernando Macedo',
    author_email='fgmacedo@gmail.com',
    description='Library to extract data from semi-structured text documents',
    long_description=long_description,
    license='MIT',
    url="http://github.org/fgmacedo/raspador",
    version='0.2.1',
    packages=['raspador'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
